import numpy as np
import matplotlib.pyplot as plt
from scipy.ndimage import gaussian_filter as gf
from scipy.interpolate import interp1d

def Add_Ellipse(cov2, center, ax, fill=False):
	det = np.linalg.det(cov2)

	W = cov2[0,0]+cov2[1,1]
	a2 = 0.5*( W + np.sqrt(W*W - 4*det) )
	b2 = 0.5*( W - np.sqrt(W*W - 4*det) )

	a = np.sqrt(a2) ; b = np.sqrt(b2)
	sin_theta = ( cov2[0,0] - a2 )
	cos_theta = cov2[0,1]
	angle = -np.arctan2( sin_theta , cos_theta )

	def get_xy(t,a,b):
		xe = center[0] + a*np.cos(t)*np.cos(angle) - b*np.sin(t)*np.sin(angle)
		ye = center[1] + a*np.cos(t)*np.sin(angle) + b*np.sin(t)*np.cos(angle)	
		return xe, ye

	if(fill):
		t = np.linspace(0,2*np.pi,100)
		x1, y1 = get_xy(t,a,b)
		x2, y2 = get_xy(t,2*a,2*b)
		ax.plot(x1,y1,'k-',lw=2,zorder=3)
		ax.plot(x2,y2,'k-',lw=2,zorder=3)
	else:
		t = np.linspace(0,2*np.pi,100)
		x1, y1 = get_xy(t,a,b)
		x2, y2 = get_xy(t,2*a,2*b)
		ax.plot(x1,y1,'r--',zorder=3)
		ax.plot(x2,y2,'r--',zorder=3)

	pass

#--------------------------------------------
# Corner Plot
#--------------------------------------------

def Curves_CL(x,y,CLs):
	h, xx, yy = np.histogram2d(x,y,bins=60)
	dx = np.average(np.diff(xx))
	dy = np.average(np.diff(yy))
	h  = gf(h,2).T ; h /= np.matrix.sum(np.matrix(h)*dx*dy)

	h2 = np.sort(h.ravel()) # matrix(n,n) to 1d-array(n*n)
	sums = np.zeros(len(h2))

	for i in range(len(h2)): sums[i] = np.sum(h2[i:])*dx*dy
	func = interp1d(sums,h2,fill_value='extrapolate')

	return xx[1:], yy[1:], h, func(CLs)

def Plot_corner(M, truths, labels, FreeParams, Cov):
	print('\n'+"-----"*10+'\n')
	Np = len(truths)
	print(">> Plotting corner ...")
	plt.rcParams.update({'font.family':'serif'})
	fig = plt.figure(figsize=(2*Np,2*Np))
	for i in range(Np): # columns
		ax = plt.subplot(Np,Np,i*Np+i+1)

		Xi = M[:,i]
		qx = [0.5*(1-0.99),0.5*(1+0.99)]
		li = np.quantile(Xi,qx)

		CL = 0.68
		qs = [0.5*(1-CL),0.5,0.5*(1+CL)]
		q = np.quantile(M[:,i],qs)
		ax.set_title('$%.2f^{+%.2e}_{-%.2e}$' % (q[1],q[1]-q[0], q[2]-q[1]))

		if('d_L' in labels[i]):
			print('-----> DL in params!!!')
			err = 0.5*(q[2]-q[0])
			err *= 100/q[1] 
			ax.plot(np.nan,np.nan,'ko',label="$\Delta d_L$: %.2e" % err + '%')

		deltaY = 0.5*(q[2]-q[0])
		ly1, ly2 = q[1]-3*deltaY, q[1]+3*deltaY

		x, bins = np.histogram(M[:,i],bins=60)
		y = gf(x,2)
		plt.fill_between(bins[1:],np.zeros(len(x)),y,color='k',alpha=0.4)
		plt.plot(bins[1:],y,'k-',lw=2)
		ax.set_xticklabels([])
		ax.set_yticklabels([])
		ax.set_xlim(ly1,ly2)
		ax.set_ylim(0,1.2*max(y))
		ax.vlines(truths[i],0,1.2*max(y),color='r')
		
		ax.vlines(q[0],0,1.2*max(y),color='k',ls='--')
		ax.vlines(q[1],0,1.2*max(y),color='k',ls='--')
		ax.vlines(q[2],0,1.2*max(y),color='k',ls='--')

		if(i==(Np-1)): ax.set_xlabel(labels[i])
		for j in range(i): # rows
			x = M[:,j]
			y = M[:,i]
			Xj = M[:,j]
			lj = np.quantile(Xj,qs)

			deltaX = 0.5*(lj[2]-lj[0])
			lx1, lx2 = lj[1]-3*deltaX, lj[1]+3*deltaX

			ax = plt.subplot(Np,Np,i*Np+j+1)

			xx, yy, h, levels = Curves_CL(x,y,[0.95,0.68,0.0])
			ax.contour(xx,yy,h,levels=levels,colors='k')
			ax.contourf(xx,yy,h,levels=levels,cmap='Greys')
			
			Cov2 = np.zeros([2,2])
			Cov2[0,0] = Cov[j,j]
			Cov2[1,1] = Cov[i,i]
			Cov2[0,1] , Cov2[1,0] = Cov[j,i], Cov[j,i]

			print("Error(%s): " % labels[j], np.sqrt(Cov2[0][0]) )
			#Add_Ellipse(Cov2,[truths[j],truths[i]],ax=ax, fill=False)

			#-----------------------------------------------------------
			ax.vlines(truths[j],ly1,ly2,color='r')
			ax.hlines(truths[i],lx1,lx2,color='r')
			ax.plot(truths[j],truths[i],'rs',mec='k')

			muY = truths[j]
			sigmaY = np.sqrt(Cov[j,j])
			
			ax.set_xlim(lx1,lx2)
			ax.set_ylim(ly1,ly2)

			ax.set_xticklabels([])
			ax.set_yticklabels([])
			if(i==(Np-1)): 	ax.set_xlabel(labels[j])
			if(j==0): ax.set_ylabel(labels[i])

	ax.plot(np.nan,np.nan,'r--',lw=2,label='From $F^{-1}$')
	print('\n'+"-----"*10+'\n')
	return fig

def gaussian(mu,sigma):
	x1 = mu-3*sigma
	x2 = mu+3*sigma
	x = np.linspace(x1,x2,100)
	y = np.exp(-(x-mu)**2/sigma**2) / np.sqrt(2*np.pi*sigma**2)
	return x, y

def Plot_Fisher(truths, labels, FreeParams, Cov):
	print('\n'+"-----"*10+'\n')
	Np = len(truths)
	print(">> Plotting Fisher ...")
	plt.rcParams.update({'font.family':'serif'})
	fig = plt.figure(figsize=(2*Np,2*Np))
	for i in range(Np): # columns
		ax = plt.subplot(Np,Np,i*Np+i+1)

		mu = truths[i]
		sigma = np.sqrt(Cov[i,i])
		x, y = gaussian(mu,sigma)

		ax.set_title('$%.2f^{+%.2e}_{-%.2e}$' % (mu,sigma,sigma))

		if('DL' in labels[i]):
			try: ax.plot(np.nan,np.nan,'ko',label="$\Delta d_L$ = %.2f" % fisher_err + '%')
			except: pass

		plt.fill_between(x,np.zeros(len(x)),y,color='k',alpha=0.4)
		plt.plot(x,y,'k-',lw=2)
		ax.set_xticklabels([])
		ax.set_yticklabels([])
		ax.set_xlim(min(x),max(x))
		ax.set_ylim(0,1.2*max(y))
		ax.vlines(truths[i],0,1.2*max(y),color='r')
		
		ax.vlines(mu-sigma,0,1.2*max(y),color='k',ls='--')
		ax.vlines(mu,0,1.2*max(y),color='k',ls='--')
		ax.vlines(mu+sigma,0,1.2*max(y),color='k',ls='--')

		if(i==(Np-1)): ax.set_xlabel(labels[i])

		muX = mu
		sigmaX = sigma
		for j in range(i): # rows
			x, f = gaussian(truths[j],np.sqrt(Cov[j,j]))
			y, f = gaussian(truths[i],np.sqrt(Cov[i,i]))

			ax = plt.subplot(Np,Np,i*Np+j+1)

			Cov2 = np.zeros([2,2])
			Cov2[0,0] = Cov[j,j]
			Cov2[1,1] = Cov[i,i]
			Cov2[0,1] , Cov2[1,0] = Cov[j,i], Cov[j,i]

			Add_Ellipse(Cov2,[truths[j],truths[i]],ax=ax, fill=True)
			
			muY = truths[j]
			sigmaY = np.sqrt(Cov[j,j])

			ax.vlines(truths[j],min(y),max(y),color='r')
			ax.hlines(truths[i],min(x),max(x),color='r')
			ax.plot(truths[j],truths[i],'rs',mec='k')

			ax.set_xlim(muX-3*sigmaX,muX+3*sigmaX)
			ax.set_ylim(muY-3*sigmaY,muY+3*sigmaY)

			ax.set_xticklabels([])
			ax.set_yticklabels([])
			if(i==(Np-1)): 	ax.set_xlabel(labels[j])
			if(j==0): ax.set_ylabel(labels[i])

	print('\n'+"-----"*10+'\n')
	return fig


from __future__ import division
import numpy as np
import bilby, os, shutil, sys
from time import time as now

import GWDALI.lib.Diff_Signal_Tensors as gwfunc
from astropy.cosmology import FlatLambdaCDM
cosmo = FlatLambdaCDM(70,0.3)

#sys.stdout = open(os.devnull, 'w')
bilby.core.utils.setup_logger(log_level=0)	# hide init bilby output in screen
bilby.core.utils.setup_logger(log_level=10)
bilby.core.utils.setup_logger(log_level=20)
bilby.core.utils.setup_logger(log_level=30)
bilby.core.utils.setup_logger(log_level=40)
bilby.core.utils.setup_logger(log_level=50)

#--------------------------------------------------
class GW_likelihood(bilby.Likelihood):
	def __init__(self,args):
		aux = {}
		self.FreeParams = args[0]
		self.Data       = args[1]
		self.GWparams   = args[2]
		self.DetAp		= args[3]
		for f in args[0]: aux[f] = None
		super().__init__(aux)

	def log_likelihood(self):
		Np = len(self.FreeParams)
		params = self.GWparams.copy()
		for fp in self.FreeParams:
			params[fp] = self.parameters[fp]
			
		loglike = 0 ; ndet = 0
		detectors, approximant = self.DetAp 

		for det in detectors:
			h = gwfunc.Signal(params, det, approximant)
			diff = self.Data[ndet]-h
			loglike -= 0.5*gwfunc.ScalarProduct(det['freq'],det['Sn'],diff,diff)
			ndet += 1
			
		if(np.isnan(loglike) or np.isinf(loglike)): loglike = -1.e9
		return loglike
#--------------------------------------------------

class DALI_likelihood(bilby.Likelihood):
	def __init__(self,args):
		self.FreeParams  = args[0]
		self.Theta0      = args[1]
		self.Tensors     = args[2]
		self.dali_method = args[3]
		aux = {}
		for f in args[0]: aux[f] = None
		super().__init__(aux)

	def log_likelihood(self):
		dX = [] ; i = 0
		Np = len(self.FreeParams)
		for fp in self.FreeParams:
			Theta  = self.parameters[fp]
			Theta0 = self.Theta0[i] ; i += 1
			dX.append(Theta-Theta0) # Fixed!!!

		Fij , Doublet , Triplet = self.Tensors
		Db3, Db4 = Doublet
		Tp4, Tp5, Tp6 = Triplet

		dX  = np.array(dX)
		dX2 = np.outer(dX,dX)
		dX3 = np.outer(dX2,dX)
		dX4 = np.outer(dX3,dX)
		dX5 = np.outer(dX4,dX)
		dX6 = np.outer(dX5,dX)

		dX2 = np.ravel(dX2)
		dX3 = np.ravel(dX3)
		dX4 = np.ravel(dX4)
		dX5 = np.ravel(dX5)
		dX6 = np.ravel(dX6)

		# From arXiv:2203.02670
		# logL = logL0 - (1/2)Fisher * dTheta_ij - [ (1/2)Db3 * dTheta_ijk  + (1/8) * dTheta_ijkl ]
		#        - [ (1/6)Tp4 * dTheta_ijkl + (1/12) * dTheta_ijklm + (1/72) * dTheta_ijklmn ] + ...
		
		loglike = -np.sum(Fij*dX2)/2
		loglike -= np.sum(Db3*dX3)/2 + np.sum(Db4*dX4)/8
		loglike -= np.sum(Tp4*dX4)/6 + np.sum(Tp5*dX5)/12 + np.sum(Tp6*dX6)/72

		if(np.isnan(loglike) or np.isnan(loglike)): return -1.e9
		else: return loglike

#----------------------------------------
# Prior(DL): dVc/dz
#---------------------------------------- 
z      = np.linspace(1.e-3,10,1000) ; dz = z[1]-z[0]# redshift
Vc     = cosmo.comoving_volume(z).value
priorD = np.diff(Vc)/np.diff(z)
z   = z[1:]-dz/2
XdL = cosmo.luminosity_distance(z).value / 1.e3 # dL in Gpc
YdL = priorD/sum(priorD)
#----------------------------------------
d1 = cosmo.luminosity_distance(1.e-3).value / 1.e3 # Gpc
d2 = cosmo.luminosity_distance(5.0).value / 1.e3   # Gpc
#----------------------------------------
# Pior(Dec): cos(Dec[degrees])
Xdec = np.linspace(-90,90,1000)
Ydec = np.cos(Xdec*np.pi/180)
Ydec/=sum(Ydec)
#----------------------------------------

def get_posterior(FreeParams, Theta0, Detection_Dict, GwData, approximant, Detectors, Tensors, dali_method, sampler_method, npoints,new_priors):
	Priors_std = {}
	Priors_std['DL']   = bilby.core.prior.Interped(name='DL',xx=XdL,yy=YdL,minimum=d1, maximum=d2) # Gpc
	Priors_std['iota']  = bilby.core.prior.Sine(name='iota', minimum=0, maximum=np.pi) # radians
	Priors_std['psi']   = bilby.core.prior.Uniform(name='psi',minimum=0, maximum=np.pi) # radians
	#----------------------------#----------------------------#----------------------------
	Priors_std['alpha'] = bilby.core.prior.Uniform(name='alpha',minimum=-np.pi, maximum=np.pi) # radians
	Priors_std['beta']  = bilby.core.prior.Sine(name='beta',minimum=0, maximum=np.pi) # radians
	#----------------------------#----------------------------#----------------------------
	Priors_std['RA']    = bilby.core.prior.Uniform(name='RA',minimum=-180, maximum=180) # degrees
	Priors_std['Dec']   = bilby.core.prior.Interped(name='Dec',xx=Xdec,yy=Ydec,minimum=-90, maximum=90) # degrees
	#----------------------------#----------------------------#----------------------------
	Priors_std['m1']    = bilby.core.prior.Uniform(name='m1',minimum=0.1, maximum=100) # solar mass
	Priors_std['m2']    = bilby.core.prior.Uniform(name='m2',minimum=0.1, maximum=100) # solar mass
	#----------------------------#----------------------------#----------------------------
	Priors_std['Mc']    = bilby.core.prior.Uniform(name='Mc',minimum=0.1, maximum=100) # solar mass
	Priors_std['eta']   = bilby.core.prior.Uniform(name='eta',minimum=1.e-3, maximum=1./4)
	Priors_std['q']     = bilby.core.prior.Uniform(name='q',minimum=1.e-3, maximum=1.0)
	#----------------------------#----------------------------#----------------------------
	Priors_std['sx1']    = bilby.core.prior.Uniform(name='sx1',minimum=0, maximum=1.0)
	Priors_std['sy1']    = bilby.core.prior.Uniform(name='sy1',minimum=0, maximum=1.0)
	Priors_std['sz1']    = bilby.core.prior.Uniform(name='sz1',minimum=0, maximum=1.0)
	Priors_std['sx2']    = bilby.core.prior.Uniform(name='sx2',minimum=0, maximum=1.0)
	Priors_std['sy2']    = bilby.core.prior.Uniform(name='sy2',minimum=0, maximum=1.0)
	Priors_std['sz2']    = bilby.core.prior.Uniform(name='sz2',minimum=0, maximum=1.0)
	#----------------------------#----------------------------#----------------------------
	Priors_std['phi_coal']  = bilby.core.prior.Uniform(name='phi_coal',minimum=0, maximum=2*np.pi) # radians
	Priors_std['t_coal']    = bilby.core.prior.Uniform(name='t_coal',minimum=0, maximum=3600) # seconds
	#----------------------------#----------------------------#----------------------------

	Priors = {}
	for fp in FreeParams: Priors[fp] = Priors_std[fp]
	if(new_priors!=None):
		for fp in new_priors.keys():
			try:
				Xval, Yval = new_priors[fp]
				Yval /= sum(Yval)
				Priors[fp] = bilby.core.prior.Interped(name=fp,xx=Xval,yy=Yval,minimum=min(Xval),maximum=max(Xval))
			except:
				print('\n>> Invalid key/argment in new_priors dictionay!\n')
				quit()

	if(dali_method=='Standard'):
		likelihood = GW_likelihood([FreeParams, GwData, Detection_Dict, [Detectors,approximant] ])
	else:
		Fisher, Doublet, Triplet = Tensors
		Fisher = np.ravel(Fisher)
		Doublet = [np.ravel(Doublet[i]) for i in range(2)]
		Triplet = [np.ravel(Triplet[i]) for i in range(3)]
		Tensors = [Fisher,Doublet,Triplet]
		likelihood = DALI_likelihood([FreeParams, Theta0, Tensors, dali_method])

	outdir = 'outdir_%s_%d/' % (dali_method, int(np.random.uniform(0,300))  )
	if(os.path.isdir(outdir)):
		shutil.rmtree(outdir)

	#----------------------------
	# Running Sampler
	#----------------------------
	print("\t >> Method:", dali_method)
	res = bilby.run_sampler(likelihood=likelihood, priors=Priors,
							sampler=sampler_method, npoints=npoints, 
							nsteps=npoints, nwalkers=npoints, 
							nburn=int(0.2*npoints), outdir=outdir)
	shutil.rmtree(outdir)

	#----------------------------
	M = [res.posterior[fp] for fp in FreeParams]
	return np.transpose(np.array(M))



import numpy as np
import matplotlib.pyplot as plt
import os

files = ['Sn_L.txt','Sn_V.txt','Sn_K.txt','Sn_ET.txt','Sn_CE.txt']
fig = plt.figure()

labels = {}
labels['Sn_L'] = 'aLIGO'
labels['Sn_V'] = 'aVirgo'
labels['Sn_K'] = 'KAGRA'
labels['Sn_ET'] = 'ET'
labels['Sn_CE'] = 'CE'

ls = {}
ls['Sn_L'] = '-'
ls['Sn_V'] = '--'
ls['Sn_K'] = ':'
ls['Sn_ET'] = '-'
ls['Sn_CE'] = '--'

colors = {}
colors['Sn_L'] = 'b'
colors['Sn_V'] = 'b'
colors['Sn_K'] = 'b'
colors['Sn_ET'] = 'r'
colors['Sn_CE'] = 'r'

for f in files:
	if(f.endswith('.txt')):
		data = np.loadtxt(f)
		x = data[:,0]
		y = np.sqrt(x)*data[:,1]
		ref = f.replace('.txt','')

		plt.loglog(x,y,color=colors[ref],ls=ls[ref],lw=3,label=labels[ref])
		print(f, len(x))

plt.legend()
plt.grid(which='both',alpha=0.3)
plt.xlabel('frequency [Hz]')
plt.ylabel('Characteristic Strain $\sqrt{f\cdot S_n(f)}$')

fig.savefig('Sn.png')
fig.savefig('Sn.pdf')

plt.show()




import numpy as np
import matplotlib.pyplot as plt
import os, shutil, sys
import warnings ; warnings.filterwarnings("ignore")
#===========================
#----Internal_libs----------
#===========================
import GWDALI.lib.Waveforms as wf
import GWDALI.lib.Angles_lib as geo
import GWDALI.lib.Dictionaries as gwdict
import GWDALI.lib.Diff_Signal_Tensors as gwfunc
import GWDALI.lib.Corner_Plots as corner
import GWDALI.lib.Auxiliar as Aux
from .lib.Likelihood import get_posterior
from .lib.Get_Tensors import Get_Tensors
#===========================
from scipy.interpolate import interp1d
from time import time as now
from tqdm import trange

#-------------------------------------------------
# From Dictionaries.py
PSD, labels_tex = gwdict.Load_Dictionaries()

#-------------------------------------------------------------------------------
#	"Compute the (Moore-Penrose) pseudo-inverse of a matrix"
#-------------------------------------------------------------------------------
def invert_matrix(matrix, rcond):

    dm = np.sqrt(np.diag(matrix))
    norm = np.outer(dm, dm)
    M_norm = matrix / norm

    try: M_inv = np.linalg.pinv(M_norm,rcond=rcond) / norm
    except: M_inv = np.linalg.pinv(matrix,rcond=rcond)

    return M_inv
#-------------------------------------------------------
# Get Waveforms and SNR
#-------------------------------------------------------
def get_Sn(det):
	name = det['name']
	try:
		Sn0 = PSD[name]
		freq0 = 10**np.linspace(0,4,4000)
		func_Sn = interp1d(freq0, Sn0, fill_value=np.inf, bounds_error=False)
	except:
		try:
			Sn0 = det['Sn']
			freq0 = det['freq']
		except:
			print("Fail on loading detectors!")
			quit()
	return Sn0, freq0

def get_strain_snr(gw_prms,detectors,approximant='TaylorF2',fmin=1.,fmax=1.e4,fsize=3.e3):
	Strains, SNR = [], []
	rho2 = 0

	freq = 10**np.linspace(np.log10(fmin), np.log10(fmax), int(fsize))
	for idx in range(len(detectors)):
		det = detectors[idx]
		Sn0, freq0 = get_Sn(det)
		func_Sn = interp1d(freq0,Sn0,fill_value=np.inf,bounds_error=False)
		detectors[idx]['freq'] = freq
		detectors[idx]['Sn']   = func_Sn(freq)

		h 	  = gwfunc.Signal(gw_prms, det, approximant)
		SNR2  = gwfunc.ScalarProduct(det['freq'], det['Sn'],h,h)
		rho2 += SNR2
		#print('#####'*10)
		Strains.append(h)
		SNR.append(np.sqrt(SNR2))

	hp, hx, freq = gwfunc.GW_Polarizations(gw_prms,freq, approximant)

	return [hp,hx,freq], Strains, SNR, np.sqrt(rho2)
#-------------------------------------------------------
# Main Function
#-------------------------------------------------------
def GWDALI( Detection_Dict, 
			FreeParams, 
			detectors, 
			approximant = 'TaylorF2_py',
			fmin  = 1., 
			fmax  = 1.e4, 
			fsize = 3000, 
			dali_method    = 'Doublet',
			sampler_method = 'nestle', 
			npoints      = 100,
			rcond		 = 1.e-4,
			new_priors   = None,
			save_samples = False, 
			save_cov     = False, 
			save_fisher  = False,
			plot_corner  = False,
			hide_info    = False,
			step_size	 = 1.e-6,
			diff_order	 = 2,
			run_sampler  = True,
			index		 = 1):

	path = Aux.CheckPath(save_cov, save_samples, save_fisher, plot_corner,dali_method)
	if(not hide_info):
		print('\n'+"-----"*10)
		print("-------------- Running GWDALI --------------------")
		print("-----"*10+'\n')
		print(" >> FreeParams: " ,  FreeParams)
		print(" >> dali_method: %s" % dali_method)
		print(" >> approximant: %s" % approximant)
		print(" >> sampler_method: %s\n" % sampler_method)

	Np   = len(FreeParams)
	keys = Detection_Dict.keys()
	freq = 10**np.linspace(np.log10(fmin), np.log10(fmax), fsize)

	#-------------------------------------------------
	# Setting Sn(f) by interpolation
	for idx in range(len(detectors)):
		det = detectors[idx]
		Sn0, freq0 = get_Sn(det)
		func_Sn = interp1d(freq0,Sn0,fill_value=np.inf,bounds_error=False)
		detectors[idx]['freq'] = freq
		detectors[idx]['Sn']   = func_Sn(freq)
	#-------------------------------------------------------
	# Store the injection values (Dict)
	truths = {}
	for key in Detection_Dict.keys():
		truths[key] = Detection_Dict[key]
	#-------------------------------------------------------
	# Store the injection values (List)
	Theta0 = np.array( [Detection_Dict[fp] for fp in FreeParams] )
	#-----------------------------------------------------------------------------------
	# From Compute_Tensors.py (looping in detectors)
	#-----------------------------------------------------------------------------------
	t1 = now() # initial time
	SNR , GwData, Fisher , Doublet , Triplet = Get_Tensors(FreeParams, Detection_Dict, detectors, dali_method, approximant, step_size, diff_order, hide_info)
	Tensors = [Fisher , Doublet , Triplet]
	t2 = now() # final time
	dT = t2 - t1
	dT_min = dT // 60
	dT_sec = dT % 60
	#-----------------------------------------------------------------------------------
	if(not hide_info): print("\n>> Fisher/Doublet/Triplet computed in %d sec" %dT)
	#------------------------------------------------------
	# Invert Fisher Matrix
	#------------------------------------------------------
	try: Cov = invert_matrix(Fisher, rcond) # from numpy.linalg.pinv
	except: Cov = np.ones([Np,Np])*np.nan
	
	Result = {}
	Result['Fisher']    = Fisher
	Result['CovFisher'] = Cov

	if(not hide_info):
		print("\n Fisher = \n", np.matrix(Fisher),'\n')
		print("\n Covariance = \n", np.matrix(Cov),'\n')
		print("-----"*10+'\n')

	if(save_fisher): Aux.Save_fisher(path,index,Detection_Dict,FreeParams,Fisher)
	if(save_cov): Aux.Save_CovFish(path,index,Detection_Dict,FreeParams,Cov)	
	#------------------------------------------------------
	# Monte-Carlo From Likelihood.py
	#------------------------------------------------------
	vec = [ truths[fp] for fp in FreeParams]
	truths = vec
	Cov2, M, Error, Recovery = [],[],[],[]
	if(dali_method != 'Fisher'):
		# allowed only for Fisher_Sampling and Doublet
		if(run_sampler):
			if(not hide_info): print("Running Sampler ...\n")
			M = get_posterior(FreeParams, Theta0, Detection_Dict, GwData, approximant, detectors, Tensors, dali_method, sampler_method, npoints, new_priors)
			qs = [0.2,0.5,0.8] # CL = 60%
			for m in M.T:
				q = np.quantile(m,qs)
				Recovery.append(q[1])
				Error.append(0.5*(q[2]-q[0]))
			Cov2 = np.matrix( np.cov(np.transpose(M)) )
			if(save_cov): Aux.Save_Cov(path,index,Detection_Dict,FreeParams,M,Cov2)
				
		Result['Covariance'] = Cov2
		Result['Samples']    = M
		Result['Error']      = Error
		Result['Recovery']   = Recovery
	
	Result['SNR']	  = SNR
	Result['Tensors'] = Tensors

	if(save_samples): Aux.Save_Samples(path,index,Detection_Dict,FreeParams,M)
	if(plot_corner): Aux.PlotCorner(M,truths,Cov,FreeParams,[dT_min,dT_sec],dali_method,path,index)

	# Result (dict)
	# keys: Fisher, CovFisher, Covariance, Samples, Error, Recovery, SNR
	return Result 


import os
import numpy as np
import matplotlib.pyplot as plt
from itertools import permutations
import GWDALI.lib.Dictionaries as gwdict
import GWDALI.lib.Corner_Plots as corner

line = "---"*30
PSD, labels_tex = gwdict.Load_Dictionaries()

def Load_Header(Detection_Dict, FreeParams):
	hd1 = 'FullInjections:\n'
	hd2 = ''
	keys = Detection_Dict.keys()
	for key in keys:
		hd1 += "%s\t" % key
		hd2 += "%e\t" % Detection_Dict[key]
	header1 = 'Injections: ' ; header2 = '' ; line = '---'*10
	for fp in FreeParams:
		header1 += '%s\t' % fp
		header2 += '%e\t' % Detection_Dict[fp]
	header = hd1 + '\n' + hd2 + '\n' + line + '\n' + header1+'\n'+header2 +'\n'+ line
	return header

def CheckPath(save_cov, save_samples, save_fisher, plot_corner,dali_method):
	cond_out = any([save_cov, save_samples, save_fisher, plot_corner])
	path = None
	if(cond_out):
		path = 'output_%s/' % dali_method
		if(not os.path.isdir(path)): os.mkdir(path)
	return path

def Save_fisher(path,index,Detection_Dict,FreeParams,Fisher):
	header = Load_Header(Detection_Dict, FreeParams) 
	np.savetxt(path+'Fisher_Matrix_%d.txt'%(index),Fisher,fmt='%e',delimiter='\t',header=header)
	pass

def Save_CovFish(path,index,Detection_Dict,FreeParams,Cov):
	header = Load_Header(Detection_Dict, FreeParams)	
	np.savetxt(path+'FisherCov_%d.txt'%(index),Cov,fmt='%e',delimiter='\t',header=header)
	pass

def Save_Cov(path,index,Detection_Dict,FreeParams,M,Cov2):
	header = Load_Header(Detection_Dict, FreeParams)
	header3 = header + '\n' + line + '\nRecovery:\n'
	for i in range(len(FreeParams)): header3 += '%e\t' % np.average(M[:,i])
	header3 += '\n' + line
	np.savetxt(path+'Covariance_Matrix_%d.txt'%(index),Cov2,fmt='%e',delimiter='\t',header=header3)
	pass

def Save_Samples(path,index,Detection_Dict,FreeParams,M):
	header = Load_Header(Detection_Dict, FreeParams)
	np.savetxt(path+'samples_%d.txt' % (index),M,fmt='%e',delimiter='\t',header=header)

def PlotCorner(M,truths,Cov,FreeParams,dT,dali_method,path,index):
	dT_min, dT_sec = dT
	labels = [ labels_tex[fp] for fp in FreeParams]
	title = "Concluded in [%d min : %d sec]" % (dT_min, dT_sec)
	if(dali_method == 'Fisher'):
		fig = corner.Plot_Fisher(truths, labels, FreeParams, Cov)
		plt.suptitle(title)
		fig.subplots_adjust(wspace=0.01, hspace=0.01)
		fig.savefig(path+'corner_%d.png' % (index))
	else:
		fig = corner.Plot_corner(M, truths, labels, FreeParams, Cov)
		plt.suptitle(title)
		fig.legend(loc='upper right', bbox_to_anchor=(0.95, 0.95))
		fig.subplots_adjust(wspace=0.01, hspace=0.01)
		fig.savefig(path+'corner_%d.png' % (index))	
	plt.show()
	pass

#===========================================#===========================================
#===========================================#===========================================

def get_indep_indexies(Np):
	#---------------------#---------------------
	# --------------Doublet(1,2)--------------
	#---------------------#---------------------
	indepD_12 = []
	for i in range(Np):
		for j in range(Np):
			for k in range(j,Np):
				indepD_12.append([i,j,k])
	#---------------------#---------------------
	# --------------Doublet(2,2)--------------
	#---------------------#---------------------
	indepD_22 = []
	for i in range(Np):
		for j in range(i,Np):
			for k in range(Np):
				for l in range(k,Np):
					indepD_22.append([i,j,k,l])
	#---------------------#---------------------
	# --------------Triplet(1,3)--------------
	#---------------------#---------------------
	indepT_13 = []
	for i in range(Np):
		for j in range(Np):
			for k in range(j,Np):
				for l in range(k,Np):
					indepT_13.append([i,j,k,l])
	#---------------------#---------------------
	# --------------Triplet(2,3)--------------
	#---------------------#---------------------
	indepT_23 = []
	for i in range(Np):
		for j in range(i,Np):
			for k in range(Np):
				for l in range(k,Np):
					for m in range(l,Np):
						indepT_23.append([i,j,k,l,m])
	#---------------------#---------------------
	# --------------Triplet(3,3)--------------
	#---------------------#---------------------
	indepT_33 = []
	for i in range(Np):
		for j in range(i,Np):
			for k in range(j,Np):
				for l in range(Np):
					for m in range(l,Np):
						for n in range(m,Np):
							indepT_33.append([i,j,k,l,m,n])
	#-----------------------------#-----------------------------
	idxs_doub = [indepD_12, indepD_22]
	idxs_trip = [indepT_13, indepT_23, indepT_33]

	return idxs_doub, idxs_trip